from .meta import APIMetadata, SerializerAsOutputMetadata
from .permissions import BlockSchemaAccess
from .router import APIRouter


__all__ = [
    "APIMetadata",
    "SerializerAsOutputMetadata",
    "BlockSchemaAccess",
    "APIRouter",
]



from typing import Any, Dict

from django.urls import NoReverseMatch
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework.views import APIView


__all__ = ["APIRootView"]


class APIRootView(APIView):
    """Welcome! This is the API root."""

    api_root_dict: Dict[str, str] = {}

    def get(self, request: Request, *args: Any, **kwargs: Any) -> Response:
        routes = {}
        namespace = request.resolver_match.namespace
        for key, url_name in self.api_root_dict.items() or {}:
            if namespace:
                url_name = namespace + ":" + url_name
            try:
                routes[key] = reverse(
                    viewname=url_name,
                    args=args,
                    kwargs=kwargs,
                    request=request,
                    format=kwargs.get("format", None),
                )
            except NoReverseMatch:
                continue

        return Response(routes)



from typing import Protocol, Set, Union

from django.utils.encoding import force_str
from rest_framework.fields import DictField, Field, SerializerMethodField
from rest_framework.metadata import SimpleMetadata
from rest_framework.request import Request, clone_request
from rest_framework.serializers import HiddenField, ListSerializer, ManyRelatedField, RelatedField, Serializer


class _View(Protocol):
    allowed_methods: Set[str]

    def get_serializer(self, *args, **kwargs) -> Serializer:
        """Get serializer"""


__all__ = ["APIMetadata", "SerializerAsOutputMetadata"]


class APIMetadata(SimpleMetadata):
    """Metadata class that adds input and output info for each of the attached views methods based on
    the serializer for that method."""

    recognized_methods = {"GET", "POST", "PUT", "PATCH", "DELETE"}
    skip_fields = [HiddenField, SerializerMethodField]

    @staticmethod
    def get_input_serializer(view: _View):
        return view.get_serializer()

    @staticmethod
    def get_output_serializer(view: _View):
        return view.get_serializer(output=True)

    def determine_actions(self, request: Request, view: _View):
        """Return information about the fields that are accepted for methods in self.recognized_methods."""
        actions = {}
        for method in self.recognized_methods & set(view.allowed_methods):
            view.request = clone_request(request, method)

            input_serializer = self.get_input_serializer(view)
            output_serializer = self.get_output_serializer(view)

            exceptions = getattr(input_serializer, "error_codes", {})
            exceptions.update(getattr(output_serializer, "error_codes", {}))

            actions[method] = {
                "input": self.get_serializer_info(input_serializer),
                "output": self.get_serializer_info(output_serializer),
                "exceptions": exceptions,
            }
            view.request = request

        return actions

    def get_serializer_info(self, serializer: Serializer) -> Union[list, dict]:
        """Given an instance of a serializer, return a dictionary of metadata about its fields."""
        data_serializer = getattr(serializer, "child", serializer)

        output_metadata = getattr(data_serializer, "output_metadata", None)
        if output_metadata is not None:
            return output_metadata

        input_data = {
            field_name: self.get_field_info(field)
            for field_name, field in data_serializer.fields.items()
            if not any(isinstance(field, field_type) for field_type in self.skip_fields)
        }

        if isinstance(serializer, ListSerializer):
            input_data = [input_data]

        return input_data

    def get_field_info(self, field: Union[Field, Serializer, ListSerializer]) -> Union[list, dict]:
        if getattr(field, "child", False):
            if isinstance(field, DictField):
                return {"<key>": self.get_field_info(field.child)}
            return [self.get_field_info(field.child)]
        if getattr(field, "fields", False):
            return self.get_serializer_info(field)

        field_info = {"type": self.label_lookup[field], "required": getattr(field, "required", False)}

        attrs = [
            "label",
            "help_text",
            "min_length",
            "max_length",
            "min_value",
            "max_value",
        ]

        for attr in attrs:
            value = getattr(field, attr, None)
            if value is not None and value != "":
                info = force_str(value, strings_only=True)
                if attr == "label" and info.lower() == field.field_name.lower().replace("_", " "):
                    continue
                field_info[attr] = info

        if (
            not getattr(field, "read_only", False)
            and hasattr(field, "choices")
            and not isinstance(field, (RelatedField, ManyRelatedField))
        ):
            field_info["choices"] = list(field.choices)

        return field_info


class SerializerAsOutputMetadata(APIMetadata):
    """Metadata class that presumes that view serializer is used as response data with no request data."""

    def determine_actions(self, request: Request, view: _View) -> Union[list, dict]:
        """Return information about the fields that are accepted for methods in self.recognized_methods."""
        actions = {}
        for method in self.recognized_methods & set(view.allowed_methods):
            view.request = clone_request(request, method)

            input_serializer = self.get_input_serializer(view)

            actions[method] = {
                "input": {},
                "output": self.get_serializer_info(input_serializer),
                "exceptions": getattr(input_serializer, "error_codes", {}),
            }
            view.request = request

        return actions



from django.conf import settings
from rest_framework.permissions import BasePermission


__all__ = ["BlockSchemaAccess"]


class BlockSchemaAccess(BasePermission):
    """Block schema access from OPTIONS request when not in DEBUG mode."""

    def has_permission(self, request, view):
        return not (request.method == "OPTIONS" and not settings.DEBUG)



from typing import Any, Callable, Dict, List, Type, Union

from django.urls import include, path, re_path
from django.urls.resolvers import URLPattern, URLResolver
from rest_framework.routers import DefaultRouter
from rest_framework.schemas import DefaultSchema
from rest_framework.urlpatterns import format_suffix_patterns
from rest_framework.views import APIView
from rest_framework.viewsets import ViewSet, ViewSetMixin

from .views import APIRootView


__all__ = ["APIRouter"]


UrlsType = list[Union[URLResolver, URLPattern]]


class APIRouter(DefaultRouter):
    """Router that will show APIViews in API root."""

    def __init__(self, name: str = None, docstring: str = None, **kwargs):
        self._navigation_routes: Dict[str, "APIRouter"] = {}

        name = name if name is not None else "APIRootView"
        self._root_view: Type[APIRootView] = type(name, (APIRootView,), {})  # noqa
        self._root_view.__doc__ = docstring if docstring else APIRootView.__doc__
        self._root_view.schema = DefaultSchema() if kwargs.get("show_in_shema", False) else None
        self._root_view._ignore_model_permissions = kwargs.get("ignore_model_permissions", False)

        self.root_view_name: str = name
        super().__init__(**kwargs)

    @property
    def navigation_routes(self) -> Dict[str, "APIRouter"]:
        """Add urls from these routers to this routers urls under the root-view of the added router,
        which will be named after the given key. This enables browser navigation of the API."""
        return self._navigation_routes

    @navigation_routes.setter
    def navigation_routes(self, value: Dict[str, "APIRouter"]):
        self._navigation_routes = value

    def get_routes(self, viewset: type[Union[ViewSet, APIView]]):
        if issubclass(viewset, ViewSetMixin):
            return super().get_routes(viewset)
        return []

    def get_api_root_view(self, api_urls: UrlsType = None) -> Callable[..., Any]:
        api_root_dict = {}
        list_name = self.routes[0].name

        for prefix, viewset, basename in self.registry:
            if issubclass(viewset, ViewSetMixin):
                api_root_dict[prefix] = list_name.format(basename=basename)
            else:
                api_root_dict[prefix] = basename

        for basename in self.navigation_routes:
            api_root_dict[fr"{basename}"] = basename

        return self._root_view.as_view(api_root_dict=api_root_dict)

    def get_urls(self) -> UrlsType:
        urls: List[Union[URLResolver, URLPattern]] = []
        for prefix, view, basename in self.registry:
            if issubclass(view, ViewSetMixin):
                continue

            regex = r"^{prefix}{trailing_slash}$".format(prefix=prefix, trailing_slash=self.trailing_slash)

            urls.append(re_path(regex, view.as_view(), name=basename))

        urls = format_suffix_patterns(urls)

        for basename, router in self.navigation_routes.items():
            router.root_view_name = basename
            urls.append(path(f"{basename}/", include(router.urls)))

        return super().get_urls() + urls



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['browsable_router']

package_data = \
{'': ['*']}

install_requires = \
['Django>=1.10',
 'djangorestframework>=3.7.0',
 'typing-extensions>=3.10.0,<4.0.0']

setup_kwargs = {
    'name': 'django-browsable-router',
    'version': '0.1.0',
    'description': 'A Django REST Framework router that can show APIViews and include other routers as navigable urls in the root view.',
    'long_description': '# Django Browsable Router\n\n[![Coverage Status](https://coveralls.io/repos/github/MrThearMan/django-browsable-router/badge.svg?branch=main)](https://coveralls.io/github/MrThearMan/django-browsable-router?branch=main)\n[![GitHub Workflow Status](https://img.shields.io/github/workflow/status/MrThearMan/django-browsable-router/Tests)](https://github.com/MrThearMan/django-browsable-router/actions/workflows/main.yml)\n[![PyPI](https://img.shields.io/pypi/v/django-browsable-router)](https://pypi.org/project/django-browsable-router)\n[![GitHub](https://img.shields.io/github/license/MrThearMan/django-browsable-router)](https://github.com/MrThearMan/django-browsable-router/blob/main/LICENSE)\n[![GitHub last commit](https://img.shields.io/github/last-commit/MrThearMan/django-browsable-router)](https://github.com/MrThearMan/django-browsable-router/commits/main)\n[![GitHub issues](https://img.shields.io/github/issues-raw/MrThearMan/django-browsable-router)](https://github.com/MrThearMan/django-browsable-router/issues)\n\n```shell\npip install django-browsable-router\n```\n---\n\n**Documentation**: [https://mrthearman.github.io/django-browsable-router/](https://mrthearman.github.io/django-browsable-router/)\n\n**Source Code**: [https://github.com/MrThearMan/django-browsable-router](https://github.com/MrThearMan/django-browsable-router)\n\n---\n',
    'author': 'Matti Lamppu',
    'author_email': 'lamppu.matti.akseli@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MrThearMan/django-browsable-router',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4',
}


setup(**setup_kwargs)
